"""
In this test we adopt the PluginTestCase provided by aiida and migrate
existing AiiDATestCase tests with minimum effort
"""


import io
import os

from aiida.common.folders import SandboxFolder
from aiida.common.exceptions import ValidationError
from unittest import TestCase
import pytest


Ti_otfg = "Ti 3|1.8|9|10|11|30U:40:31:32(qc=5.5)"
Sr_otfg = "Sr 3|2.0|5|6|7|40U:50:41:42"
O_otfg = "O 2|1.1|15|18|20|20:21(qc=7)"

# This fixture will be run for each class and ensure
# AiiDA test environment has been loaded correctly


@pytest.fixture(scope="class")
def import_things(aiida_profile, request):
    import aiida_castep.data.otfg as otf
    import aiida_castep.data.usp as usp
    from aiida.orm import DataFactory
    request.cls.otfg = DataFactory("castep.otfgdata")
    request.cls.otf = otf
    request.cls.usp = usp
    request.cls.usp = usp


@pytest.mark.usefixtures("import_things")
class BaseDataCase(TestCase):
    """Base to include some useful things"""

    # @pytest.fixture(autouse=True, scope="function")
    # def populate_env(self):
    #     """
    #     Popoulate the class namespace with the imported modules
    #     """
    #     if not self.AIIDA_ENV_LOADED:
    #         import aiida_castep.data.otfg as otf
    #         import aiida_castep.data.usp as usp
    #         from aiida.orm import DataFactory
    #         self.otfg = DataFactory("castep.otfgdata")
    #         self.otf = otf
    #         self.usp = usp
    #         self.AIIDA_ENV_LOADED = True


    @pytest.fixture(autouse=True)
    def reset_db(self, aiida_profile):
        aiida_profile.reset_db()
        yield
        aiida_profile.reset_db()

    def create_family(self):
        """Creat families for testsing"""
        Ti, Sr, O, C9 = self.get_otfgs()
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, O.entry], "STO_FULL", "TEST", False)
        self.otf.upload_otfg_family([Ti.entry, Sr.entry], "STO_O_missing", "TEST", False)

        # Missing O but that's OK we have a C9 wild card here
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, "C9"], "STO_O_C9", "TEST", False)

        return "STO_FULL", "STO_O_missing", "STO_O_C9"

    def get_otfgs(self):

        Ti, _ = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        Sr, _ = self.otfg.get_or_create(Sr_otfg, store_otfg=False)
        O, _ = self.otfg.get_or_create(O_otfg, store_otfg=False)
        C9, _ = self.otfg.get_or_create("C9", store_otfg=False)
        return Ti, Sr, O, C9

    @staticmethod
    def get_STO_structure():
        """Return a STO structure"""
        from aiida.orm import DataFactory
        StructureData = DataFactory("structure")
        a = 3.905

        cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols=["Sr"])
        s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["Ti"])
        s.append_atom(position=(a / 2, a / 2, 0.), symbols=["O"])
        s.append_atom(position=(a / 2, 0., a / 2), symbols=["O"])
        s.append_atom(position=(0., a / 2, a / 2), symbols=["O"])
        s.label = "STO"
        return s


class TestOTFGData(BaseDataCase):


    def setUp(self):
        self.otfg_nodes = {}

    def test_otfg_split(self):
        element, setting = self.otf.split_otfg_entry(Ti_otfg)
        self.assertEqual(element, "Ti")
        self.assertEqual(setting, "3|1.8|9|10|11|30U:40:31:32(qc=5.5)")

    def test_otfg_create(self):

        element, setting = self.otf.split_otfg_entry(Ti_otfg)
        C9 = self.otfg(string="C9")
        self.assertEqual(C9.string, "C9")
        self.assertEqual(C9.element, "LIBRARY")
        C9.store()

        Ti = self.otfg(string=setting, element=element)
        self.assertEqual(Ti.string, setting)
        self.assertEqual(Ti.element, element)
        Ti.store()

    def test_get_or_create(self):

        Ti, create = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        self.assertTrue(create)

        # Create but not stored do it again should have no change
        Ti, create = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        self.assertTrue(create)

        Ti.store()

        Ti2, create = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        self.assertFalse(create)

        # Should get the stored OTFG entry
        # If using SQLA this Ti should be IS Ti2?
        self.assertEqual(Ti2.uuid, Ti.uuid)

        # Using another way should be the same
        Sr = self.otfg()
        Sr.set_element("Sr")
        Sr.set_string("bla")
        Sr.store()

        Sr2, create = self.otfg.get_or_create("Sr_bla")
        self.assertFalse(create)

        # A different OTFG
        Sr3, create = self.otfg.get_or_create("Sr_foo")
        self.assertTrue(create)

        # Check if more than one instance is found in the db
        with self.assertRaises(ValueError):
            Sr = self.otfg()
            Sr.set_element("Sr")
            Sr.set_string("bla")
            Sr.store()

            # This should fail there are more than one
            Sr4, create = self.otfg.get_or_create("Sr_bla", use_first=False)

        Sr4, create = self.otfg.get_or_create("Sr_bla", use_first=True)

    def test_set_up_family(self):

        Ti, Sr, O, C9 = self.get_otfgs()

        otfgs = [Ti.entry, Sr.entry, C9.entry]
        entry, uploaded = self.otf.upload_otfg_family(otfgs, "Test", "Test")
        self.assertEqual((entry, uploaded), (3, 3))

        # This should fail
        with self.assertRaises(ValidationError):
            entry, uploaded = self.otf.upload_otfg_family(otfgs, "Test", "Test")

        entry, uploaded = self.otf.upload_otfg_family([O.entry] + otfgs, "Test", "Test", stop_if_existing=False)

        groups = self.otfg.get_otfg_groups()
        self.assertEqual(len(groups), 1)

        retrieved_entries = [node.entry for node in groups[0].nodes]
        for o in otfgs + [O.entry]:
            self.assertIn(o, retrieved_entries)

    def test_assign_from_structure(self):
        """
        Test using get_pseudos_from_structure
        """

        from aiida_castep.data import get_pseudos_from_structure
        from aiida.common.exceptions import NotExistent

        self.create_family()
        STO = self.get_STO_structure()

        pseudo_list = get_pseudos_from_structure(STO, "STO_FULL")
        self.assertEqual(pseudo_list["Sr"].entry, Sr_otfg)
        self.assertEqual(pseudo_list["O"].entry, O_otfg)
        self.assertEqual(pseudo_list["Ti"].entry, Ti_otfg)

        with self.assertRaises(NotExistent):
            pseudo_list = get_pseudos_from_structure(STO, "STO_O_missing")

        pseudo_list = get_pseudos_from_structure(STO, "STO_O_C9")
        self.assertEqual(pseudo_list["Sr"].entry, Sr_otfg)
        self.assertEqual(pseudo_list["O"].entry, "C9")
        self.assertEqual(pseudo_list["Ti"].entry, Ti_otfg)

    def create_family(self):
        """Creat families for testsing"""
        Ti, Sr, O, C9 = self.get_otfgs()
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, O.entry], "STO_FULL", "TEST", False)
        self.otf.upload_otfg_family([Ti.entry, Sr.entry], "STO_O_missing", "TEST", False)

        # Missing O but that's OK we have a C9 wild card here
        self.otf.upload_otfg_family([Ti.entry, Sr.entry, "C9"], "STO_O_C9", "TEST", False)

    def get_otfgs(self):

        Ti, _ = self.otfg.get_or_create(Ti_otfg, store_otfg=False)
        Sr, _ = self.otfg.get_or_create(Sr_otfg, store_otfg=False)
        O, _ = self.otfg.get_or_create(O_otfg, store_otfg=False)
        C9, _ = self.otfg.get_or_create("C9", store_otfg=False)
        return Ti, Sr, O, C9


class TestUspData(BaseDataCase):


    def upload_usp_family(self):
        """Make a fake usp node"""

        with SandboxFolder() as f:
            sub = f.get_subfolder("pseudo", create=True)
            for element in ["Sr", "Ti", "O"]:
                fp = io.StringIO(u"foo bla 42")
                sub.create_file_from_filelike(fp, "{}_00.usp".format(element))

            self.usp.upload_usp_family(os.path.join(f.abspath, "pseudo"), "STO", "")

            with self.assertRaises(ValueError):
                self.usp.upload_usp_family(os.path.join(f.abspath, "pseudo"), "STO", "")

    def get_usp_node(self, element):
        """
        Return a node of usp file
        """
        name = "{}_00.usp".format(element)
        with SandboxFolder() as f:
            fp = io.StringIO(u"foo bla 42")
            f.create_file_from_filelike(fp, name)
            fpath = os.path.join(f.abspath, name)
            node = self.usp.UspData.get_or_create(fpath)[0]

        return node

    def test_get_or_create(self):
        """Testing the logic or get_or_create"""
        name = "Sr_00.usp"
        with SandboxFolder() as f:
            fp = io.StringIO(u"foo bla 42")
            f.create_file_from_filelike(fp, name)
            fpath = os.path.join(f.abspath, name)
            node1, create = self.usp.UspData.get_or_create(fpath)

            self.assertTrue(create)
            self.assertEqual(node1.element, "Sr")

            node2 = self.usp.UspData(file=fpath)
            node2.store()

            # Now having two files - should raise an exception
            with self.assertRaises(ValueError):
                node3, create = self.usp.UspData.get_or_create(fpath, use_first=False)

            # This should work now
            node4, create = self.usp.UspData.get_or_create(fpath, use_first=True)
            self.assertFalse(create)
            self.assertIn(node4.pk, (node1.pk, node2.pk))

    def test_upload(self):
        self.upload_usp_family()

    def test_assign_from_structure(self):
        """
        Test using get_pseudos_from_structure
        """

        from aiida_castep.data import get_pseudos_from_structure
        from aiida.common.exceptions import NotExistent

        self.upload_usp_family()
        STO = self.get_STO_structure()

        pseudo_list = get_pseudos_from_structure(STO, "STO")
        for kind in STO.kinds:
            self.assertIn(kind.name, pseudo_list)

        with self.assertRaises(NotExistent):
            STO.append_atom(symbols="Ba", position=(1, 1, 1))
            pseudo_list = get_pseudos_from_structure(STO, "STO")


if __name__ == "__main__":
    import unittest
    from aiida.utils.fixtures import TestRunner, BACKEND_SQLA
    tests = unittest.defaultTestLoader.discover(".", pattern="dbtestdata_new.py")
    #TestRunner().run(tests)
    TestRunner().run(tests, backend=BACKEND_SQLA)



"""
Test parsing data
"""
from __future__ import print_function
from aiida_castep.parsers.raw_parser import parse_castep_text_output, parse_geom_text_output, parse_dot_bands
import unittest
import pytest

import os

@pytest.fixture(scope="class")
def import_things(aiida_profile, request):

    from aiida.orm import DataFactory, CalculationFactory
    FolderData = DataFactory("folder")
    for k, v in locals().items():
        setattr(request.cls, k, v)

class TestParsers(unittest.TestCase):

    @property
    def data_abs_path(self):
        test_moudule = os.path.split(__file__)[0]
        data_folder = os.path.join(test_moudule, "data")
        return data_folder

    def setUp(self):
        with open(self.data_abs_path + "/H2-geom/aiida.geom") as fh:
            self.geom_lines = fh.readlines()

        with open(self.data_abs_path + "/H2-geom/aiida.castep") as cs:
            self.castep_lines = cs.readlines()
        pass

    def test_parse_geom(self):
        res = parse_geom_text_output(self.geom_lines, None)
        self.assertEqual(res["symbols"], ["H", "H"])
        self.assertEqual(res["geom_total_energy"].shape[0], 5)
        self.assertEqual(res["positions"].shape[0], 5)
        self.assertEqual(res["forces"].shape[0], 5)

    def test_parse_castep(self):
        parsed_data, trajectory_data, warnings = parse_castep_text_output(self.castep_lines, None)
        self.assertTrue(parsed_data["cell_constraints"])
        self.assertFalse(parsed_data["warnings"])
        self.assertTrue(parsed_data["n_kpoints"])
        self.assertTrue(parsed_data["space_group"])
        self.assertTrue(parsed_data["pseudo_pots"])
        self.assertTrue(parsed_data["point_group"])
        self.assertTrue(trajectory_data["enthalpy"])
        self.assertTrue(trajectory_data["total_energy"])
        self.assertLess(trajectory_data["total_energy"][0] - trajectory_data["enthalpy"][0], 1e-5)
        self.assertFalse(parsed_data["warnings"])
        self.assertEqual(parsed_data["total_time"], 14.53)
        self.assertEqual(parsed_data["initialisation_time"], 1.02)
        self.assertEqual(parsed_data["parallel_efficiency"], 90)
        self.assertEqual(parsed_data["castep_version"], "17.2")

    def test_warnings(self):
        # Test assertion of warnings
        with_warning = self.castep_lines[:]
        # This is no longer a critical warning leading to FAILED state
        with_warning.insert(-10, "Geometry optimization failed to converge")
        parsed_data, trajectory_data, critical = parse_castep_text_output(with_warning, None)
        self.assertTrue(parsed_data["warnings"])
        self.assertNotIn(parsed_data["warnings"][0], critical)

        with_warning = self.castep_lines[:]
        with_warning.insert(-10, "SCF cycles performed but system has not reached the groundstate")
        parsed_data, trajectory_data, critical = parse_castep_text_output(with_warning, None)
        self.assertTrue(parsed_data["warnings"])
        self.assertIn(parsed_data["warnings"][0], critical)

    def test_parse_bands(self):
        """
        Test the function to parse *.bands file
        """
        res = parse_dot_bands(os.path.join(self.data_abs_path, "Si-geom-stress/aiida.bands"))
        self.assertEqual(res[0]['nspins'], 1)
        self.assertEqual(res[0]['nkpts'], len(res[1]))
        self.assertEqual(res[0]['neigns'], len(res[2][0][0]))

    def test_parser_stress(self):
        with open(self.data_abs_path + "/Si-geom-stress/aiida.castep") as cs:
            lines = cs.readlines()
        parsed_data, trajectory_data, warnings = parse_castep_text_output(lines, None)

        self.assertIn('symm_stress', trajectory_data)
        self.assertIn('symm_pressure', trajectory_data)
        self.assertTrue(trajectory_data['symm_pressure'])
        print(trajectory_data['symm_pressure'])



@pytest.mark.usefixtures("import_things", "aiida_profile")
class TestCastepParser(object):

    def get_data_abs_path(self):
        test_moudule = os.path.split(__file__)[0]
        data_folder = os.path.join(test_moudule, "data")
        return data_folder

    def get_dummy_outputs(self):

        retrieved = {}
        for folder in ["H2-geom", "O2-geom-spin", "Si-geom-stress", "N2-md"]:
            folderdata = self.FolderData()
            folderdata.replace_with_folder(
                os.path.join(self.get_data_abs_path(), folder))
            retrieved[folder] = dict(retrieved=folderdata)
        return retrieved

    def test_version_consistency(self):
        """
        Test that moudule versions are consistent
        """
        from aiida_castep.parsers.raw_parser import __version__ as prv
        from aiida_castep.parsers.castep import __version__ as pcv
        from aiida_castep.calculations.castep import __version__ as ccv
        from aiida_castep.calculations.base import __version__ as cbv
        assert prv == pcv == ccv == cbv

    def test_parser_retrieved(self, code_echo, localhost, STO_calculation):
        from .utils import get_x2_structure
        from aiida_castep.parsers.castep import CastepParser

        calc = STO_calculation

        parser = CastepParser(calc)
        retrived_folders = self.get_dummy_outputs()
        common_keys = ["cells", "positions", "forces", "symbols", "geom_total_energy"]
        md_keys = ["hamilt_energy", "kinetic_energy",
                   "velocities", "temperatures", "times"]
        geom_keys = ["geom_enthalpy"]

        for name, r in retrived_folders.items():
            if 'O2' in name:
                xtemp = 'O'
            elif 'Si' in name:
                xtemp = 'Si'
            elif 'N2' in name:
                xtemp = 'N'
            elif 'H2' in name:
                xtemp = 'H'
            # Swap the correct structure to allow desort to work
            calc.use_structure(get_x2_structure(xtemp))

            success, out = parser.parse_with_retrieved(r)
            out = dict(out)

            out_structure = out[parser.get_linkname_outstructure()]
            out_param_dict = out[parser.get_linkname_outparams()].get_dict()
            out_traj = out[parser.get_linkname_outtrajectory()]
            assert "total_energy" in  out_param_dict
            assert "unit_energy" in  out_param_dict
            assert out_param_dict["unit_energy"] == "eV"
            # Check if the label is correctly copied
            assert calc.inp.structure.label == out_structure.label

            # Check the length of sites are consistent
            assert len(out_structure.sites) == len(out_traj.get_symbols())

            for k in common_keys:
                assert k in out_traj.get_arraynames()

            if name == "O2-geom-spin" or name == "Si-geom-stress":
                assert parser.get_linkname_outbands() in out
                bands = out[parser.get_linkname_outbands()]

                # Check if spins are handled correctly
                assert bands.get_attr('nspins') in [1, 2]
                if bands.get_attr('nspins') == 1:
                    assert bands.get_attr('nkpts') == len(bands.get_bands())
                elif bands.get_attr('nspins') == 2:
                    assert bands.get_attr('nkpts') == len(bands.get_bands()[0])

                for k in geom_keys:
                    assert k in out_traj.get_arraynames()

            if name == "N2-md":
                for k in md_keys:
                    assert k in out_traj.get_arraynames()


class TestPot1DParser(object):

    def test_load_plugin(self):
        from aiida.parsers import ParserFactory
        Pot1dParser = ParserFactory("castep.pot1d")



"""
Parsers for CASTEP
"""
from aiida.orm import DataFactory
from aiida.parsers.parser import Parser  # , ParserParamManager
from aiida_castep.parsers.raw_parser import parse_raw_ouput, units
from aiida_castep.parsers.raw_parser import __version__ as raw_parser_version
from aiida_castep.parsers.utils import (structure_from_input, add_last_if_exists,
                                        desort_structure, get_desort_args)
from .._version import calc_parser_version
__version__ = calc_parser_version

ParameterData = DataFactory("parameter")
BandsData = DataFactory("array.bands")

ERR_FILE_WARNING_MSG = ".err files found in workdir"

# No need to have consistent raw parser version
#assert __version__ == raw_parser_version, "Inconsistent version numbers"


class CastepParser(Parser):
    """
    This is the class for Parsing results from a CASTEP calculation
    Supported calculations types:
    signlepoint
    geom
    """

    _setting_key = 'parser_options'

    def __init__(self, calc):
        """
        Initialise the instance of CastepParser
        """

        super(CastepParser, self).__init__(calc)

    def parse_with_retrieved(self, retrieved):
        """
        Receives a dictionary of retrieved nodes.retrieved.
        Top level logic of operation
        """

        import os

        successful = True
        seed_name = self._calc._SEED_NAME
        warnings = []

        # Look for lags of the parser
        try:
            parser_opts = self._calc.inp.settings.get_dict()[
                self.get_parser_settings_key()]
        except (AttributeError, KeyError):
            parser_opts = {}

        # NOT READLLY IN USE

        input_dict = self._calc.inp.parameters.get_dict()

        # Check that the retrieved folder is there
        try:
            out_folder = retrieved[self._calc._get_linkname_retrieved()]
        except KeyError:
            self.logger.error("No retrieved folder found")
            return False, ()

        # check what is inside the folder
        list_of_files = out_folder.get_folder_list()

        # at least the stdout should exist
        if self._calc._OUTPUT_FILE_NAME not in list_of_files:
            self.logger.error("Standard output not found")
            successful = False
            return successful, ()

        # The calculation is failed if there is any err file.
        for f in list_of_files:
            if ".err" in f:
                successful = False
                self.logger.warning("Error files found in workdir.")
                warnings.append(ERR_FILE_WARNING_MSG)
                break

        # look for other files

        # Trajectory files
        has_md_geom = False
        if seed_name + ".geom" in list_of_files:
            out_md_geom_file = os.path.join(
                out_folder.get_abs_path('.'), seed_name + '.geom')
            has_md_geom = True
        elif seed_name + ".md" in list_of_files:
            out_md_geom_file = os.path.join(
                out_folder.get_abs_path('.'), seed_name + '.md')
            has_md_geom = True
        else:
            out_md_geom_file = None
            has_md_geom = False

        # Handling bands
        if self._calc._SEED_NAME + ".bands" in list_of_files:
            has_bands = True
            out_bands_file = os.path.join(out_folder.get_abs_path('.'),
                                          seed_name + '.bands')
        else:
            has_bands = False
            out_bands_file = None

        out_file = os.path.join(out_folder.get_abs_path(
            '.'), self._calc._OUTPUT_FILE_NAME)

        # call the raw parsing function
        parsing_args = [out_file, input_dict,
                        parser_opts, out_md_geom_file,
                        out_bands_file]

        # If there is a geom file then we parse it
        out_dict, trajectory_data, structure_data, bands_data, raw_sucessful\
            = parse_raw_ouput(*parsing_args)

        # Append the final value of trajectory_data into out_dict
        last_value_keys = ["free_energy", "total_energy",
                           "zero_K_energy", "spin_density",
                           "abs_spin_density", "enthalpy"]
        for key in last_value_keys:
            add_last_if_exists(trajectory_data, key, out_dict)

        # Add warnings from this level
        out_dict["warnings"].extend(warnings)

        successful = all([raw_sucessful, successful])

        # Saving to nodes
        new_nodes_list = []

        ######## --- PROCESSING BANDS DATA -- ########
        if has_bands:
            bands = bands_to_bandsdata(bands_data)
            new_nodes_list.append((self.get_linkname_outbands(), bands))

        ######## --- PROCESSING STRUCTURE DATA --- ########
        try:
            cell = structure_data["cell"]
            positions = structure_data["positions"]
            symbols = structure_data["symbols"]

        except KeyError:
            # No final structure can be used - that is OK
            pass
        else:
            output_structure = structure_from_input(
                cell=cell, positions=positions, symbols=symbols)
            calc_in = self._calc
            # Use the output label as the input label
            input_structure = calc_in.get_inputs_dict()[calc_in.get_linkname("structure")]
            output_structure = desort_structure(output_structure, input_structure)
            output_structure.label = input_structure.label
            new_nodes_list.append(
                (self.get_linkname_outstructure(), output_structure))

        ######### --- PROCESSING TRAJECTORY DATA --- ########
        # If there is anything to save
        # It should...
        if trajectory_data:

            import numpy as np
            from aiida.orm.data.array.trajectory import TrajectoryData
            from aiida.orm.data.array import ArrayData

            # If we have .geom file, save as in a trajectory data
            if has_md_geom:
                try:
                    idesort = get_desort_args(input_structure)
                    positions = np.asarray(trajectory_data["positions"])[:, idesort]
                    cells = trajectory_data["cells"]
                    symbols = np.asarray(trajectory_data["symbols"])[idesort]
                    stepids = np.arange(len(positions))

                except KeyError:
                    out_dict["parser_warning"].append("Cannot "
                                                      "extract data from .geom file.")
                    pass

                else:
                    traj = TrajectoryData()
                    traj.set_trajectory(stepids=np.asarray(stepids),
                                        cells=np.asarray(cells),
                                        symbols=np.asarray(symbols),
                                        positions=np.asarray(positions))
                    # Save the rest
                    for name, value in trajectory_data.iteritems():
                        # Skip saving empty arrays
                        if len(value) > 0:
                            traj.set_array(name, np.asarray(value))
                    new_nodes_list.append(
                        (self.get_linkname_outtrajectory(), traj))

            # Otherwise, save data into a ArrayData node
            else:
                out_array = ArrayData()
                for name, value in trajectory_data.iteritems():
                    # Skip saving empty arrays
                    if len(value) > 0:
                        out_array.set_array(name, np.asarray(value))
                new_nodes_list.append((self.get_linkname_outarray(),
                                       out_array))

        ######## ---- PROCESSING OUTPUT DATA --- ########
        output_params = ParameterData(dict=out_dict)
        new_nodes_list.append((self.get_linkname_outparams(), output_params))
        return successful, new_nodes_list

    # getter method for various names
    @classmethod
    def get_parser_settings_key(cls):
        """
        Returns the name of the key to be used in the calculation settings, that
        contains the dictionary with the parser_options.
        Not used for now
        """
        return 'parser_options'

    @classmethod
    def get_linkname_outstructure(cls):
        """
        Returns the name of the link to the output_structure
        Only exists if it is a geometry optimisation run.
        """
        return 'output_structure'

    @classmethod
    def get_linkname_outtrajectory(cls):
        """
        Returns the name of the link to the output_trajectory.
        Node exists in case of calculation = "geometryoptimsiation"
        """
        return 'output_trajectory'

    @classmethod
    def get_linkname_outarray(cls):
        """
        Returns the name of the link to the output_array.
        Exist if trajectory data cannot be created e.g not a optimisation run.
        """
        return 'output_array'

    @classmethod
    def get_linkname_out_kpoints(cls):
        """
        Not implemented for now
        """
        return 'output_kpoints'

    @classmethod
    def get_linkname_outbands(cls):
        """
        Returns the name of the link to the output band data.
        Exists if we retrieved the bands file
        """
        return 'output_bands'


class Pot1dParser(Parser):
    """
    Parser for Pot1d
    """

    def parse_with_retrieved(self, retrieved):
        # NOT READLLY IN USE

        # Check that the retrieved folder is there
        try:
            out_folder = retrieved[self._calc._get_linkname_retrieved()]
        except KeyError:
            self.logger.error("No retrieved folder found")
            return False, ()

        # check what is inside the folder
        list_of_files = out_folder.get_folder_list()

        # at least the stdout should exist
        oname = self._calc._OUTPUT_FILE_NAME
        if oname not in list_of_files:
            self.logger.error("Standard output not found")
            successful = False
            return successful, ()

        # The calculation is failed if there is any err file.
        for f in list_of_files:
            if ".err" in f:
                successful = False
                self.logger.warning("Error files found in workdir.")
                break

        # Check for keyword
        castep_file = out_folder.get_file_content(self._calc._OUTPUT_FILE_NAME)
        if "Finished pot1d" in castep_file:
            successful = True
        else:
            successful = False

        return successful, []


def bands_to_bandsdata(bands_res):
    """
    Convert the result of parser_dot_bands into a BandsData object

    :param bands_res: Output from ``parse_dot_bands`` function.
    :return: A BandsData object
    :rtype: ``aiida.orm.bands.data.array.bands.BandsData``
    """

    import numpy as np
    bands = BandsData()

    # Extract the index of the kpoints
    # kpn_array are rows (kindex, kx, ky, kz, weight)
    kpn_array = np.array(bands_res[1])
    k_index = kpn_array[:, 0]

    # We need to restore the order of the kpoints
    k_sort = np.argsort(k_index)
    # Sort the kpn_array
    kpn_array = kpn_array[k_sort]

    _weights = kpn_array[:, -1]
    kpts = kpn_array[:, 1:-1]
    bands.set_kpoints(kpts, weights=_weights)

    # We need to swap the axes from kpt,spin,engs to spin,kpt,engs
    bands_array = np.array(bands_res[2])[k_sort]  #  Sort the bands
    bands_array = bands_array.swapaxes(0, 1)

    # Squeeze the first dimension e.g when there is a single spin
    if bands_array.shape[0] == 1:
        bands_array = bands_array[0]
    bands_array = bands_array * units['Eh']
    bands_res[0]['efermi'] *= units['Eh']
    bands_res[0]['units'] = "eV"

    bands.set_bands(bands_array)
    bands.set_cell(bands_res[0]['cell'], pbc=(True, True, True))

    # Store information from *.bands in the attributes
    # This is needs as we need to know the number of electrons
    # and the fermi energy
    for key, value in bands_res[0].items():
        bands._set_attr(key, value)
    return bands



"""
Module for storing usp files into the database
We take many things from upf.py here as usp is not too much different...
TODO: Add function for usp file validation and parsing
"""

import re
import os
from aiida.orm import DataFactory
from aiida.common.utils import classproperty

USPGROUP_TYPE = "data.castep.usp.family"
SinglefileData = DataFactory("singlefile")

# Extract element from filename
re_fn = re.compile("^([a-z]+)_\w+\.(usp|recpot)$", flags=re.IGNORECASE)


def upload_usp_family(folder,
                      group_name,
                      group_description,
                      stop_if_existing=True):
    """
    Upload a set of usp/recpot files in a give group

    :param folder: a path containing all UPF files to be added.
        Only files ending in .usp/.recpot are considered.
    :param group_name: the name of the group to create. If it exists and is
        non-empty, a UniquenessError is raised.
    :param group_description: a string to be set as the group description.
        Overwrites previous descriptions, if the group was existing.
    :param stop_if_existing: if True, check for the md5 of the files and,
        if the file already exists in the DB, raises a MultipleObjectsError.
        If False, simply adds the existing UPFData node to the group.
    """
    import os

    import aiida.common
    from aiida.common import aiidalogger
    from aiida.orm import Group
    from aiida.common.exceptions import UniquenessError, NotExistent
    from aiida.orm.querybuilder import QueryBuilder

    files = [
        os.path.realpath(os.path.join(folder, i)) for i in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, i)) and (
            i.lower().endswith('.usp') or i.lower().endswith('recpot')
            or i.lower().endswith('.uspcc'))
    ]

    nfiles = len(files)

    try:
        group = Group.get(name=group_name, type_string=USPGROUP_TYPE)
        group_created = False
    except NotExistent:
        group = Group(
            name=group_name,
            type_string=USPGROUP_TYPE,
        )
        group_created = True

    # Update the descript even if the group already existed
    group.description = group_description

    pseudo_and_created = []

    for f in files:

        md5sum = aiida.common.utils.md5_file(f)
        qb = QueryBuilder()
        qb.append(UspData, filters={'attributes.md5': {'==': md5sum}})
        existing_usp = qb.first()

        # Add the file if it is in the database
        if existing_usp is None:
            pseudo, created = UspData.get_or_create(
                f, use_first=True, store_usp=False)
            pseudo_and_created.append((pseudo, created))

        # The same file is there already
        else:
            if stop_if_existing:
                raise ValueError("A usp/recpot with identical MD5 to"
                                 " {} cannot be added with stop_if_existing"
                                 "".format(f))
            existing_usp = existing_usp[0]
            pseudo_and_created.append((existing_usp, False))

    # Check for unique per element
    elements = [(i[0].element, i[0].md5sum) for i in pseudo_and_created]

    # Check if we will duplicate after insertion

    if not group_created:
        for aiida_n in group.nodes:
            if not isinstance(aiida_n, UspData):
                continue
            elements.append((aiida_n.element, aiida_n.md5sum))

    # Discard duplicated pairs
    elements = set(elements)
    elements_names = [e[0] for e in elements]

    # Check the uniqueness of the complete group
    if not len(elements_names) == len(set(elements_names)):
        duplicates = set(
            [x for x in elements_names if elements_names.count(x) > 1])
        dup_string = ", ".join(duplicates)
        raise UniquenessError(
            "More than one usp/recpot found for the elements: " + dup_string +
            ".")

    if group_created:
        group.store()

    # Save the usp in the database if necessary and add them to the group

    for pseudo, created in pseudo_and_created:
        if created:
            pseudo.store()
            aiidalogger.debug("New node {} created for file {}".format(
                pseudo.uuid, pseudo.filename))
        else:
            aiidalogger.debug("Reusing node {} for file {}".format(
                pseudo.uuid, pseudo.filename))

    group.add_nodes(pseduo for pseduo, _ in pseudo_and_created)

    nuploaded = len([_ for _, created in pseudo_and_created if created])

    return nfiles, nuploaded


class UspData(SinglefileData):
    """
    Class for a single usp file
    These usp files are stored as individual file nodes in the database
    """

    @classmethod
    def get_or_create(cls, filename, use_first=False, store_usp=True):
        """
        Same ase init. Check md5 in the db, it is found return a UspData.
        Otherwise will store the data into the db

        :return (usp, created)
        """

        import aiida.common.utils
        import os

        # Convert the filename to an absolute path
        if filename != os.path.abspath(filename):
            raise ValueError("filename must be an absolute path")
        md5 = aiida.common.utils.md5_file(filename)

        # Check if we have got the file already
        pseudos = cls.from_md5(md5)
        if len(pseudos) == 0:
            # Not stored
            if store_usp is True:
                instance = cls(file=filename).store()
                return (instance, True)
            else:
                instance = cls(file=filename)
                return (instance, True)
        else:
            if len(pseudos) > 1:
                if use_first:
                    return (pseudos[0], False)
                else:
                    pks = ", ".join([str(i.pk) for i in pseudos])
                    raise ValueError("More than one copy of a pseudopotential"
                                     " found. pks={}".format(pks))
            else:
                return (pseudos[0], False)

    @classmethod
    def from_md5(cls, md5):
        """
        Return a list of all usp pseudopotentials that match a given MD5 hash.

        Note that the hash has to be stored in a md5 attribute, otherwise
        the pseudo will not be found.
        We use a special md5 attribute to avoid searching through
        irrelevant data types.
        """
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(cls, filters={'attributes.md5': {'==': md5}})
        return [_ for [_] in qb.all()]

    @classproperty
    def uspfamily_type_string(cls):
        return USPGROUP_TYPE

    def store(self, *args, **kwargs):
        """
        Store the node. Automatically set md5 and element
        """
        self._validate()

        return super(UspData, self).store(*args, **kwargs)

    def set_file(self, filename):
        """Added file but also check data"""
        from aiida.common.exceptions import ParsingError
        import aiida.common.utils

        try:
            element = get_usp_element(filename)
        except KeyError:
            raise ParsingError(
                "Cannot extract element form the usp/recpot file"
                " {}; Cannot store into database."
                .format(self.filename))
        md5sum = aiida.common.utils.md5_file(filename)

        super(UspData, self).set_file(filename)

        self._set_attr('element', str(element))  # For unicode/ascii ?
        self._set_attr('md5', md5sum)

    @property
    def element(self):
        return self.get_attr('element', None)

    @property
    def md5sum(self):
        """MD5 sum of the usp/recpot file"""
        return self.get_attr('md5', None)

    @classmethod
    def get_usp_group(cls, group_name):
        """
        Return the UspFamily group with the given name.
        """
        from aiida.orm import Group

        return Group.get(
            name=group_name, type_string=cls.uspfamily_type_string)

    @classmethod
    def get_usp_groups(cls, filter_elements=None, user=None):
        """
        Return all names of groups of type UspFamily,
        possibly with some filters.

        :param filter_elements: A string or a list of strings.
               If present, returns only the groups that contains one Usp for
               every element present in the list. Default=None, meaning that
               all families are returned.
        :param user: if None (default), return the groups for all users.
               If defined, it should be either a DbUser instance, or a string
               for the username (that is, the user email).
        """

        from aiida.orm import Group

        group_query_params = {"type_string": cls.uspfamily_type_string}

        if user is not None:
            group_query_params['user'] = user

        if isinstance(filter_elements, basestring):
            filter_elements = [filter_elements]

        if filter_elements is not None:
            actual_filter_elements = {_.capitalize() for _ in filter_elements}

            group_query_params['node_attributes'] = {
                'element': actual_filter_elements
            }

        all_usp_groups = Group.query(**group_query_params)

        groups = [(g.name, g) for g in all_usp_groups]
        # Sort by name
        groups.sort()
        # Return the groups, without name
        return [_[1] for _ in groups]

    def _validate(self):
        from aiida.common.exceptions import ValidationError
        import aiida.common.utils

        super(UspData, self)._validate()

        usp_abspath = self.get_file_abs_path()

        if not usp_abspath:
            raise ValidationError("No valid usp file was passed")

        element = get_usp_element(usp_abspath)
        md5 = aiida.common.utils.md5_file(usp_abspath)

        if element is None:
            raise ValidationError(
                "Cannot infer element."
                "File should be in the format '<element>_<label>.usp'."
            )

        try:
            attr_element = self.get_attr('element')
        except AttributeError:
            raise ValidationError("attribute 'element' not set.")

        try:
            attr_md5 = self.get_attr('md5')
        except AttributeError:
            raise ValidationError("attribute 'md5' not set.")

        if md5 != attr_md5:
            raise ValidationError(
                "Mismatch between store md5 and actual md5 value")

        # Check matching of data and actual file
        if attr_element != element:
            raise ValidationError("Attribute 'element' says '{}' but '{}' was "
                                  "parsed instead.".format(
                                      attr_element, element))


def get_usp_element(filepath):
    """
    infer element from usp/recpot filename
    :return element: a string of element name or None if not found
    """

    filename = os.path.split(filepath)[1]
    match = re_fn.match(filename)
    if match:
        element = match.group(1)
        element = element.capitalize()
        return element



"""
Tests for calculation module
"""
import pytest


def test_castep_summary(STO_calculation):
    # Test the get_castep_input_summary method

    keys = ["kpoints", "structure", "code", "computer", "resources",
            "custom_scheduler_commands", "wallclock", "label", "pseudos"]
    out_dict = STO_calculation.get_castep_input_summary()
    for k in keys:
        assert k in out_dict

    # Store the node, this should not change anything
    STO_calculation.store_all()
    out_dict = STO_calculation.get_castep_input_summary()
    for k in keys:
        assert k in out_dict


def test_update_parameters(STO_calculation):
    """
    Test the update_parameters method
    """

    sto = STO_calculation
    updates = {"task": "geometryoptimisation",
               "xc_functional": "pbe",
               "fix_all_cell": True}
    sto.update_parameters(**updates)
    dtmp = sto.inp.parameters.get_dict()
    assert dtmp["PARAM"]["task"] == updates["task"]
    assert dtmp["PARAM"]["xc_functional"] == updates["xc_functional"]
    assert dtmp["CELL"]["fix_all_cell"] == updates["fix_all_cell"]

    sto.update_parameters(delete=["task"])
    assert "task" not in dtmp["PARAM"]

    sto.inp.parameters.store()
    with pytest.raises(RuntimeError):
        sto.update_parameters(delete=["task"])

    # Unlink the parameters
    sto._remove_link_from(sto.get_linkname("parameters"))
    # This should still work, a new input ParameterData is created
    sto.update_parameters(**updates)
    assert sto.get_linkname("parameters") in \
        sto.get_inputs_dict()
